import numpy as np
import pandas as pd
import scanpy as sc
import matplotlib.pyplot as plt
import networkx as nx


def graph_prune(adata, group_key, emb_key, ncells_thre=0.5, weight_thre=0.6):
    '''
    :param adata: anndata to use
    :param group_key: cell cluster key to use
    :param emb_key: basis/embedding key to use
    :param ncells_thre: threshold of percent for number of cells
    :param weight_thre: threshold of weight for cluster edges which are kept
    :param split_weight: threshold of weight for cluster edges which are deleted
    :return: dataframe information for selected clusters and edges; weight matrix for selected clusters and edges
    '''
    sc.tl.diffmap(adata)
    sc.tl.paga(adata, groups=group_key)

    embedding = adata.obsm[emb_key]
    adata.obs["{}0".format(emb_key)] = embedding[:, 0]
    adata.obs["{}1".format(emb_key)] = embedding[:, 1]
    info = pd.concat([adata.obs.groupby(group_key)["{}0".format(emb_key)].mean(),
                      adata.obs.groupby(group_key)["{}1".format(emb_key)].mean(),
                      pd.Series(adata.obs[group_key].value_counts(), name="ncells")], axis=1)

    groups_sizes = adata.obs[group_key].value_counts()
    median_group_size = np.median(groups_sizes)
    groups_sizes = pd.Series(100 * np.power(groups_sizes / median_group_size, 0.5), name="sizes")
    info = pd.concat([info, groups_sizes], axis=1)

    try:
        colors = pd.Series(adata.uns[group_key + "_colors"], index=info.index, name="colors")
        info = pd.concat([info, colors], axis=1)
    except KeyError:
        pass

    # weight matrix from PAGA
    weight = adata.uns["paga"]["connectivities_tree"].A
    weight = weight + weight.T

    select_num = 0
    for i in range(len(info.index)):
        if info.sort_values(by=["ncells"], ascending=False).iloc[:i + 1, 2].sum() > ncells_thre * info["ncells"].sum():
            select_num = i + 1
            break

    if select_num < 5:
        select_num = min(5, len(info.index))
    select_clusters = list(info.sort_values(by=["ncells"], ascending=False).index[:select_num])

    # if a single cluster has over 10% of the total number of cells
    all_clusters = list(info.index)
    for cluster_name in all_clusters:
        pos = all_clusters.index(cluster_name)
        if info.iloc[pos, 2] > 0.1 * info["ncells"].sum():
            if cluster_name not in select_clusters:
                select_clusters.append(cluster_name)

    # if a cluster is connected to clusters just selected with weight over 0.6, also select it
    for cluster_name in select_clusters:
        pos = all_clusters.index(cluster_name)
        indices = np.where(weight[:, pos] > weight_thre * np.max(weight))[0]
        if len(indices) > 0:
            for idx in indices:
                cluster_new = all_clusters[idx]
                if cluster_new not in select_clusters:
                    select_clusters.append(cluster_new)

    select_index = info.index.isin(select_clusters)
    select_info = info.iloc[select_index, :]
    select_weight = weight[select_index, :][:, select_index]

    # connect two or more sub-graphs together
    n_clusters = select_info.shape[0]
    point_list = [i for i in range(n_clusters)]
    link_list = [(i, j) for i in range(n_clusters) for j in range(n_clusters) if select_weight[i, j] > 0 and i < j]
    G = nx.Graph()
    for node in point_list:
        G.add_node(node)
    for link in link_list:
        G.add_edge(link[0], link[1])

    ncells_list = []
    for c in nx.connected_components(G):
        ncells_list.append(select_info.iloc[list(c)]["ncells"].sum())

    sorted_graph = [list(nx.connected_components(G))[i] for i in list(np.argsort(-np.array(ncells_list)))]

    if len(sorted_graph) > 1:
        for p in range(len(sorted_graph)):
            for p_ in range(p + 1, len(sorted_graph)):
                c, c_ = sorted_graph[p], sorted_graph[p_]
                path_list, path_list_ = list(c), list(c_)

                path_list = [list(select_info.index)[i] for i in path_list]
                path_list = [list(info.index).index(i) for i in path_list]
                path_list_ = [list(select_info.index)[i] for i in path_list_]
                path_list_ = [list(info.index).index(i) for i in path_list_]

                path_list.sort()
                path_list_.sort()
                weight_dict = {}

                # select connected edge between c and c_
                for j in path_list:
                    for j_ in path_list_:
                        w = weight[j, j_]
                        weight_dict[(j, j_)] = w

                if max(weight_dict.values()) > 0:
                    for key, value in weight_dict.items():
                        if value == max(weight_dict.values()):
                            j, j_ = key
                            j = list(select_info.index).index(info.index[j])
                            j_ = list(select_info.index).index(info.index[j_])
                            select_weight[j, j_] = value

                # add another node to connect c and c_
                else:
                    not_selected = [i for i in list(info.index) if i not in select_clusters]
                    weight_dict = {}
                    for i in not_selected:
                        idx = list(info.index).index(i)
                        weight_idx = []
                        for j in path_list:
                            for j_ in path_list_:
                                w_c, w_c_ = weight[idx, j], weight[idx, j_]
                                weight_idx.append(w_c * w_c_)
                        weight_idx = np.max(weight_idx)
                        weight_dict[i] = weight_idx

                    if max(weight_dict.values()) > 0:
                        for key, value in weight_dict.items():
                            if value == max(weight_dict.values()):
                                select_clusters.append(key)
                                select_index = info.index.isin(select_clusters)
                                select_info = info.iloc[select_index, :]
                                select_weight = weight[select_index, :][:, select_index]

                    else:
                        raise ValueError("There are more then one unconnected sub-graphs in the dataset."
                                         "Please split the dataset and run TIVelo for them independently.")

    return select_info, select_weight


def visualize(select_info, select_weight, show=False):
    fig, ax = plt.subplots(1, 1)
    for i in range(select_weight.shape[0]):
        for j in range(select_weight.shape[1]):
            if select_weight[i, j] > 0:
                x = select_info["{}0".format(emb_key)][i], select_info["{}0".format(emb_key)][j]
                y = select_info["{}1".format(emb_key)][i], select_info["{}1".format(emb_key)][j]
                ax.plot(x, y, lw=5 * select_weight[i, j], c="black", zorder=0)

    try:
        ax.scatter(select_info["{}0".format(emb_key)], select_info["{}1".format(emb_key)], s=select_info['sizes'],
                   c=select_info["colors"], zorder=1)
    except KeyError:
        ax.scatter(select_info["{}0".format(emb_key)], select_info["{}1".format(emb_key)], s=select_info['sizes'],
                   zorder=1)

    for count, cluster_name in enumerate(select_info.index):
        ax.text(select_info["{}0".format(emb_key)][count], select_info["{}1".format(emb_key)][count], cluster_name,
                ha="center", zorder=2)
    if show:
        plt.show()
    else:
        return ax


if __name__ == '__main__':
    # test
    data_dir = r"D:\cuhk\project\velocity\dataset\atac\HSPCs\3423-MV-2_adata_postpro.h5ad"
    group_key = "leiden"
    emb_key = "X_umap"

    adata_rna = sc.read_h5ad(data_dir)
    select_info, select_weight = graph_prune(adata_rna, group_key, emb_key)
    ax = visualize(select_info, select_weight, show=False)
    plt.show()


import numpy as np
from .model import orient_score, visualize
from tqdm import tqdm
from joblib import Parallel, delayed
import scanpy as sc
from utils.velocity_genes import compute_velocity_genes
import os


def recover_direction(adata, order_idx, step=1, vis_idx=None):
    u_path, s_path = adata.layers["Mu"], adata.layers["Ms"]
    u_path, s_path = u_path[order_idx, :], s_path[order_idx, :]

    dict_tree = {}
    tree_dict = {}

    r = Parallel(n_jobs=-1, verbose=10)(delayed(orient_score)(u_path[:, g].squeeze(), s_path[:, g].squeeze(), step=step)
                                        for g in range(adata.n_vars))
    score_list, _, _, u_tree, s_tree, _ = zip(*r)

    for g in range(adata.n_vars):
        tree_dict['u_tree'] = u_tree[g]
        tree_dict['s_tree'] = s_tree[g]
        dict_tree[str(g)] = tree_dict.copy()

    score_list = np.array(score_list)

    mean = np.mean(score_list)
    median = np.median(score_list)
    uq = np.quantile(score_list, 0.75)
    lq = np.quantile(score_list, 0.25)
    minimum = np.min(score_list)
    maximum = np.max(score_list)
    pos_num = np.sum(score_list > 0)

    print("mean: {:.3f}".format(mean), "\nmedian: {:.3f}".format(median), "\nlower quantile: {:.3f}".format(lq),
          "\nupper quantile: {:.3f}".format(uq), "\nminimum: {:.3f}".format(minimum), "\nmaximum: {:.3f}".format(maximum),
          "\nNo. of positive scores: {}".format(pos_num))

    if vis_idx is not None:
        gene_name = adata.var_names[vis_idx]
        score, u_arr, s_arr, u_tree, s_tree, sec_list = orient_score(u_path[:, vis_idx].squeeze(),
                                                                     s_path[:, vis_idx].squeeze())

        fig, (ax1, ax2) = visualize(u_arr, s_arr, u_tree, s_tree, sec_list, fig_size=(12, 9))
        ax1.legend(loc="upper right", ncol=2, prop={'size': 12})
        ax2.legend(loc="upper right", ncol=2, prop={'size': 12})
        ax1.set_title("{}, orientation score: {:.2f}".format(gene_name, score), fontsize=20)
        fig.tight_layout()

        return score_list, dict_tree, (ax1, ax2)

    else:
        return score_list, dict_tree


def correct_path(path_dict, adata, group_key, rev_stat="mean", tree_gene=None, root_select="connectivities",
                 emb_key=None):

    # main path
    start_node = path_dict["start_node"]
    main_path = path_dict["main_path"]

    velocity_genes = compute_velocity_genes(adata, n_top_genes=2000, r2_adjust=True, inplace=False)

    # main_path
    main_path_cell = adata.obs[group_key].isin(main_path)
    adata_new = adata[main_path_cell, :]

    start_cell_idx = np.flatnonzero(adata_new.obs[group_key] == start_node)
    success_node_list = [node for node in main_path if node != start_node]
    success_cell_idx = []
    for success_node in success_node_list:
        success_cell_idx += list(np.flatnonzero(adata_new.obs[group_key] == success_node))

    if root_select == "connectivities":
        root_cell_idx = np.flatnonzero(adata_new.obs[group_key] == start_node)[0]
    elif root_select == "coordinates" and emb_key is not None:
        coordinates = adata_new.obsm[emb_key]
        start_xy = coordinates[start_cell_idx, :]
        success_xy = coordinates[success_cell_idx, :]
        dist_xy = (np.sum(np.square(start_xy), axis=1, keepdims=True) +
                   np.sum(np.square(success_xy), axis=1, keepdims=True).T - 2 * np.dot(start_xy, success_xy.T))
        distances = np.partition(dist_xy, 30, axis=1)[:, :30]
        root_cell_idx = start_cell_idx[np.argmin(np.sum(distances, axis=1))]
    elif root_select == "terminal_score":
        root_score = adata_new.obs["root_cells"]
        start_root_score = root_score[start_cell_idx]
        root_cell_idx = start_cell_idx[np.argmax(start_root_score)]
    elif root_select == "coordinates" and emb_key is None:
        raise KeyError("Embedding key not provided")
    else:
        raise KeyError("No such method for root cell selection")

    adata_new.uns['iroot'] = root_cell_idx
    sc.tl.diffmap(adata_new)
    sc.tl.dpt(adata_new)

    adata_new = adata_new[:, velocity_genes]
    t_paga = adata_new.obs["dpt_pseudotime"].values
    order_idx = np.argsort(t_paga)

    # recover direction
    print("\nmain path:", main_path)
    if tree_gene is not None and tree_gene in adata_new.var_names:
        gene_idx = list(adata_new.var_names).index(tree_gene)
        score_list, dict_all, ax = recover_direction(adata_new, order_idx, vis_idx=gene_idx)
    else:
        score_list, dict_all = recover_direction(adata_new, order_idx)
        ax = None

    adata.uns["score_main"] = score_list.copy()

    # judge if the direction should be reversed
    if rev_stat == "median":
        if np.median(score_list) < 0:
            path_dict["start_node"] = main_path[-1]
            path_dict["main_path"] = main_path[::-1]
    elif rev_stat == "mean":
        if np.mean(score_list) < 0:
            path_dict["start_node"] = main_path[-1]
            path_dict["main_path"] = main_path[::-1]
    else:
        raise KeyError("Unknown statistics for reversing main path: {}".format(rev_stat))

    # branches
    for key, branch in path_dict.items():
        if key.startswith("branch"):
            start_node = branch[0]
            branch_cell = adata.obs[group_key].isin(branch)
            adata_new = adata[branch_cell, :]

            start_cell_idx = np.flatnonzero(adata_new.obs[group_key] == start_node)
            success_node_list = [node for node in branch if node != start_node]
            success_cell_idx = []
            for success_node in success_node_list:
                success_cell_idx += list(np.flatnonzero(adata_new.obs[group_key] == success_node))

            if root_select == "connectivities":
                root_cell_idx = np.flatnonzero(adata_new.obs[group_key] == start_node)[0]
            elif root_select == "coordinates" and emb_key is not None:
                coordinates = adata_new.obsm[emb_key]
                start_xy = coordinates[start_cell_idx, :]
                success_xy = coordinates[success_cell_idx, :]
                dist_xy = (np.sum(np.square(start_xy), axis=1, keepdims=True) +
                           np.sum(np.square(success_xy), axis=1, keepdims=True).T - 2 * np.dot(start_xy, success_xy.T))
                distances = np.partition(dist_xy, 30, axis=1)[:, :30]
                root_cell_idx = start_cell_idx[np.argmin(np.sum(distances, axis=1))]
            elif root_select == "terminal_score":
                root_score = adata_new.obs["root_cells"]
                start_root_score = root_score[start_cell_idx]
                root_cell_idx = start_cell_idx[np.argmax(start_root_score)]
            elif root_select == "coordinates" and emb_key is None:
                raise KeyError("Embedding key not provided")
            else:
                raise KeyError("No such method for root cell selection")

            adata_new.uns['iroot'] = root_cell_idx
            sc.tl.diffmap(adata_new)
            sc.tl.dpt(adata_new)

            adata_new = adata_new[:, velocity_genes]
            t_paga = adata_new.obs["dpt_pseudotime"].values
            order_idx = np.argsort(t_paga)

            print("\n" + key + ":", branch)
            score_list, dict_all = recover_direction(adata_new, order_idx)
            adata.uns["score_" + key] = score_list

            if np.median(score_list) < 0 and np.mean(score_list) < 0:
                path_dict[key] = branch[::-1]

    # further correct the result according to overall result
    main_path = path_dict["main_path"]
    del_key = []

    for key, obj in path_dict.items():
        if key.startswith("branch"):
            if path_dict[key][-1] == main_path[0]:
                path_dict["main_path"] = path_dict[key] + main_path[1:]
                path_dict["start_node"] = path_dict[key][0]
                del_key.append(key)
            elif path_dict[key][0] == main_path[-1]:
                path_dict["main_path"] = main_path + path_dict[key][1:]
                path_dict["start_node"] = main_path[0]
                del_key.append(key)
            elif path_dict[key][0] != main_path[-1] and path_dict[key][-1] != main_path[0]:
                if path_dict[key][-1] in main_path:
                    path_dict[key] = path_dict[key][::-1]

    for branch in del_key:
        path_dict.pop(branch)

    adata.uns["path_dict"] = path_dict.copy()
    print("\n'path_dict' added to adata.uns")
    return path_dict, ax






import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import BatchSampler, RandomSampler
from tqdm import tqdm
import numpy as np
from utils.velocity_genes import compute_velocity_genes


class model_velo(object):
    def __init__(self, x_u, x_s, d_nn, knn):
        self.x_u = torch.from_numpy(x_u)
        self.x_s = torch.from_numpy(x_s)
        self.n_obs = self.x_u.shape[0]
        self.n_vars = self.x_u.shape[1]
        self.knn = torch.from_numpy(knn).to(torch.float32)
        self.d_nn = torch.from_numpy(d_nn).to(torch.float32)

        self.n_neighs_d = torch.sum(self.d_nn, dim=1)
        self.n_neighs_d[torch.where(self.n_neighs_d == 0)] = 1
        self.d_u = torch.mm(self.d_nn, self.x_u) / self.n_neighs_d.unsqueeze(1)
        self.d_s = torch.mm(self.d_nn, self.x_s) / self.n_neighs_d.unsqueeze(1)

        v_u, v_s = self.d_u - self.x_u, self.d_s - self.x_s
        self.v_u = nn.Parameter(v_u, requires_grad=True)
        self.v_s = nn.Parameter(v_s, requires_grad=True)

    def simple_fit(self):
        v_u, v_s = self.d_u - self.x_u, self.d_s - self.x_s
        v_u, v_s = v_u.clone().detach().numpy(), v_s.clone().detach().numpy()
        return v_u, v_s

    def fit(self, n_epochs=100, lr=0.001, alpha_1=1, alpha_2=0.1, batch_size=None, loss_fun="mse", only_s=False,
            same_c=None):
        same_c = torch.from_numpy(same_c).to(torch.float32) if same_c is not None else None
        mse = nn.MSELoss()
        cos_sim = nn.CosineSimilarity(dim=1)
        optimizer = optim.Adam([{'params': self.v_u}, {'params': self.v_s}], amsgrad=True, lr=lr)

        # compute loss
        pbar = tqdm(range(n_epochs))
        for epoch in pbar:
            if batch_size is None:
                optimizer.zero_grad()

                mse_u = mse(self.x_u + self.v_u, self.d_u)
                mse_s = mse(self.x_s + self.v_s, self.d_s)

                cos_xu = torch.mean(cos_sim(self.v_u, self.d_u - self.x_u))
                cos_xs = torch.mean(cos_sim(self.v_s, self.d_s - self.x_s))

                # cosine similarity with velocity
                norm_v_u = torch.norm(self.v_u, p=2, dim=1, keepdim=True)
                norm_v_s = torch.norm(self.v_s, p=2, dim=1, keepdim=True)

                cos_u = torch.mm(self.v_u, self.v_u.t()) / (torch.mm(norm_v_u, norm_v_u.t()) + 1e-6)
                cos_s = torch.mm(self.v_s, self.v_s.t()) / (torch.mm(norm_v_s, norm_v_s.t()) + 1e-6)

                if same_c is None:
                    n_neighs = torch.sum(self.knn, dim=1)
                    cos_u = torch.mean(torch.sum(cos_u * self.knn, dim=1) / (n_neighs + 1e-6))
                    cos_s = torch.mean(torch.sum(cos_s * self.knn, dim=1) / (n_neighs + 1e-6))
                else:
                    n_neighs = torch.sum(same_c * self.knn, dim=1)
                    cos_u = torch.mean(torch.sum(cos_u * same_c * self.knn, dim=1) / (n_neighs + 1e-6))
                    cos_s = torch.mean(torch.sum(cos_s * same_c * self.knn, dim=1) / (n_neighs + 1e-6))

                if loss_fun == "mse":
                    if not only_s:
                        loss = mse_u + alpha_1 * mse_s - alpha_2 * cos_u - alpha_2 * cos_s
                        loss.backward()
                        optimizer.step()

                        pbar.set_description("Model training")
                        pbar.set_postfix(mse_u=np.around(mse_u.clone().detach().numpy(), 3),
                                         mse_s=np.around(mse_s.clone().detach().numpy(), 3),
                                         cos_u=np.around(cos_u.clone().detach().numpy(), 3),
                                         cos_s=np.around(cos_s.clone().detach().numpy(), 3))
                    else:
                        loss = mse_s - alpha_2 * cos_s
                        loss.backward()
                        optimizer.step()

                        pbar.set_description("Model training")
                        pbar.set_postfix(mse_s=np.around(mse_s.clone().detach().numpy(), 3),
                                         cos_s=np.around(cos_s.clone().detach().numpy(), 3))
                elif loss_fun == "cos":
                    if not only_s:
                        loss = -cos_xu - alpha_1 * cos_xs - alpha_2 * cos_u - alpha_2 * cos_s
                        loss.backward()
                        optimizer.step()

                        pbar.set_description("Model training")
                        pbar.set_postfix(cos_xu=np.around(cos_xu.clone().detach().numpy(), 3),
                                         cos_xs=np.around(cos_xs.clone().detach().numpy(), 3),
                                         cos_u=np.around(cos_u.clone().detach().numpy(), 3),
                                         cos_s=np.around(cos_s.clone().detach().numpy(), 3))
                    else:
                        loss = -cos_xs - alpha_2 * cos_s
                        loss.backward()
                        optimizer.step()

                        pbar.set_description("Model training")
                        pbar.set_postfix(cos_xs=np.around(cos_xs.clone().detach().numpy(), 3),
                                         cos_s=np.around(cos_s.clone().detach().numpy(), 3))

            else:
                train_loader = list(BatchSampler(RandomSampler(range(self.n_obs)), batch_size=batch_size, drop_last=False))
                mse_u_, mse_s_, cos_xu_, cos_xs_, cos_u_, cos_s_ = [], [], [], [], [], []

                for step, indices in enumerate(train_loader):
                    optimizer.zero_grad()

                    x_u, x_s = self.x_u[indices, :], self.x_s[indices, :]
                    d_u, d_s = self.d_u[indices, :], self.d_s[indices, :]

                    mse_u = mse(x_u + self.v_u[indices, :], d_u)
                    mse_s = mse(x_s + self.v_s[indices, :], d_s)

                    cos_xu = torch.mean(cos_sim(self.v_u[indices, :], d_u - x_u))
                    cos_xs = torch.mean(cos_sim(self.v_s[indices, :], d_s - x_s))

                    norm_v_u = torch.norm(self.v_u[indices, :], p=2, dim=1, keepdim=True)
                    norm_v_s = torch.norm(self.v_s[indices, :], p=2, dim=1, keepdim=True)

                    cos_u = torch.mm(self.v_u[indices, :], self.v_u[indices, :].t()) / (torch.mm(norm_v_u, norm_v_u.t()) + 1e-6)
                    cos_s = torch.mm(self.v_s[indices, :], self.v_s[indices, :].t()) / (torch.mm(norm_v_s, norm_v_s.t()) + 1e-6)

                    n_neighs = torch.mean(torch.sum(self.knn[indices, :][:, indices], dim=1))
                    cos_u = torch.sum(torch.mean(cos_u * self.knn[indices, :][:, indices], dim=1)) / (n_neighs + 1e-6)
                    cos_s = torch.sum(torch.mean(cos_s * self.knn[indices, :][:, indices], dim=1)) / (n_neighs + 1e-6)

                    if loss_fun == "mse":
                        if not only_s:
                            loss = mse_u + alpha_1 * mse_s - alpha_2 * cos_u - alpha_2 * cos_s
                            loss.backward()
                            optimizer.step()
                        else:
                            loss = mse_s - alpha_2 * cos_s
                            loss.backward()
                            optimizer.step()

                    elif loss_fun == "cos":
                        if not only_s:
                            loss = -cos_xu - alpha_1 * cos_xs - alpha_2 * cos_u - alpha_2 * cos_s
                            loss.backward()
                            optimizer.step()
                        else:
                            loss = -cos_xs - alpha_2 * cos_s
                            loss.backward()
                            optimizer.step()
                    else:
                        raise KeyError("No such loss function for simple method")

                    mse_u_.append(mse_u.unsqueeze(0).clone().detach())
                    mse_s_.append(mse_s.unsqueeze(0).clone().detach())
                    cos_xu_.append(cos_xu.unsqueeze(0).clone().detach())
                    cos_xs_.append(cos_xs.unsqueeze(0).clone().detach())
                    cos_u_.append(cos_u.unsqueeze(0).clone().detach())
                    cos_s_.append(cos_s.unsqueeze(0).clone().detach())

                mse_u = np.mean(torch.cat(mse_u_).numpy())
                mse_s = np.mean(torch.cat(mse_s_).numpy())
                cos_xu = np.mean(torch.cat(cos_xu_).numpy())
                cos_xs = np.mean(torch.cat(cos_xs_).numpy())
                cos_u = np.mean(torch.cat(cos_u_).numpy())
                cos_s = np.mean(torch.cat(cos_s_).numpy())

                pbar.set_description("Model training")

                if loss_fun == "mse":
                    if not only_s:
                        pbar.set_postfix(mse_u=np.around(mse_u, 3), mse_s=np.around(mse_s, 3),
                                         cos_u=np.around(cos_u, 3), cos_s=np.around(cos_s, 3))
                    else:
                        pbar.set_postfix(mse_s=np.around(mse_s, 3), cos_s=np.around(cos_s, 3))
                elif loss_fun == "cos":
                    if not only_s:
                        pbar.set_postfix(cos_xu=np.around(cos_xu, 3), cos_xs=np.around(mse_s, 3),
                                         cos_u=np.around(cos_u, 3), cos_s=np.around(cos_s, 3))
                    else:
                        pbar.set_postfix(cos_xs=np.around(cos_xs, 3), cos_s=np.around(cos_s, 3))

        v_u, v_s = self.v_u.clone().detach().numpy(), self.v_s.clone().detach().numpy()
        return v_u, v_s


def get_velocity(adata, d_nn, knn, same_c=None, n_epochs=100, loss_fun="mse", only_s=False, constrain=True, alpha_1=1,
                 alpha_2=0.1, batch_size=512, filter_genes=False):

    if filter_genes:
        velocity_genes = compute_velocity_genes(adata, n_top_genes=2000, inplace=False)
        adata = adata[:, velocity_genes]

    x_u = adata.layers["Mu"]
    x_s = adata.layers["Ms"]

    model = model_velo(x_u, x_s, d_nn, knn)
    if constrain:
        v_u, v_s = model.fit(batch_size=batch_size, n_epochs=n_epochs, loss_fun=loss_fun, only_s=only_s,
                             alpha_1=alpha_1, alpha_2=alpha_2, same_c=same_c)
    else:
        v_u, v_s = model.simple_fit()

    adata.layers["velocity"] = v_s
    adata.layers["velocity_u"] = v_u

    return adata, v_u, v_s







import numpy as np
"""
Adapted from UniTVelo
Selects a set of genes to regularize in gene-space
"""


def get_weight(x, y=None, perc=95):
    from scipy.sparse import issparse

    xy_norm = np.array(x.A if issparse(x) else x)
    if y is not None:
        if issparse(y):
            y = y.A
        xy_norm = xy_norm / np.clip(np.max(xy_norm, axis=0), 1e-3, None)
        xy_norm += y / np.clip(np.max(y, axis=0), 1e-3, None)

    if isinstance(perc, int):
        weights = xy_norm >= np.percentile(xy_norm, perc, axis=0)
    else:
        lb, ub = np.percentile(xy_norm, perc, axis=0)
        weights = (xy_norm <= lb) | (xy_norm >= ub)

    return weights


def R2(residual, total):
    r2 = np.ones(residual.shape[1]) - \
         np.sum(residual * residual, axis=0) / \
         np.sum(total * total, axis=0)
    r2[np.isnan(r2)] = 0
    return r2


def compute_velocity_genes(adata_raw, n_top_genes, r2_adjust=True, inplace=True, min_ratio=0.01, min_r2=0.01,
                           max_r2=0.95, perc=[5, 95]):
    adata = adata_raw.copy()

    fit_offset = False
    vkey = 'velocity'

    Ms = adata_raw.layers['Ms'].copy()
    Mu = adata_raw.layers['Mu'].copy()
    n_obs, n_vars = Ms.shape
    n_var = n_vars

    gamma = np.zeros(n_vars)
    r2 = np.zeros(n_vars)
    velocity_genes = np.ones(n_vars)
    residual_scale = np.zeros([n_obs, n_vars])

    # need get_weight function
    weights = get_weight(Ms, Mu, perc=95)

    Ms_quantile, Mu_quantile = weights * Ms, weights * Mu

    # linear slope using quantiles (weights)
    gamma_quantile = np.sum(Mu_quantile * Ms_quantile, axis=0) / np.sum(Ms_quantile * Ms_quantile, axis=0)

    scaling = np.std(Mu, axis=0) / np.std(Ms, axis=0) if np.std(Ms, axis=0).any() != 0 else 1

    # switch back to non quantile
    if r2_adjust:
        Ms, Mu = Ms, Mu

    # non quantile slope
    gamma_ref = np.sum(Mu * Ms, axis=0) / np.sum(Ms * Ms, axis=0)
    residual_scale = Mu - gamma_ref * Ms

    r2 = R2(residual_scale, total=Mu - np.mean(Mu, axis=0))

    # select genes
    velocity_genes = np.ones(n_var)
    velocity_genes = (
            (r2 > min_r2)
            & (r2 < max_r2)
            & (gamma_quantile > min_ratio)
            & (gamma_ref > min_ratio)
            & (np.max(Ms > 0, axis=0) > 0)
            & (np.max(Mu > 0, axis=0) > 0)
    )

    # filter noisy genes
    if r2_adjust:
        lb, ub = np.nanpercentile(scaling, [10, 90])
        velocity_genes = (
                velocity_genes
                & (scaling > np.min([lb, 0.03]))
                & (scaling < np.max([ub, 3]))
        )

    nonzero_s, nonzero_u = Ms > 0, Mu > 0
    weights = np.array(nonzero_s & nonzero_u, dtype=bool)
    nobs = np.sum(weights, axis=0)

    velocity_genes = velocity_genes & (nobs > 0.05 * Ms.shape[1])

    if inplace:
        adata_raw.var['velocity_genes'] = velocity_genes
    else:
        return velocity_genes